VERSION_REQURE_FILE = 2
ENV_STORAGE_FILE = "RESPONSE_FILE"
ENV_REPLACEMENT_FILE = "REPLACEMENT_FILE"
ENV_REPLACEMENT_NAME = "REPLACEMENT_VAR"
ENV_DEBUG = "DEBUG"
ENV_APPLY_LATENCY = "LATENCY"
ENV_REQURE_STORAGE_MODE = "REQURE_MODE"
REPLACE_DEFAULT_KEY = "FILTERS"
METATADA_KEY = "_requre"
KEY_MINIMAL_MATCH = 2
RELATIVE_TEST_DATA_DIRECTORY = "test_data"
